/**
  * Cringasu Daniel
  * 332 CA
  *
  * Proiectare cu Microprocesoare
  */

#include <avr/io.h>
#include <avr/interrupt.h>
#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>	

#define SUS				PA0
#define STANGA			PA1
#define SELECT			PA2
#define DREAPTA			PA3
#define JOS				PA4
#define MOD_BUTTON		PA5
#define MOD_LED_GREEN	PA6
#define MOD_LED_RED		PA7

#define USER_LED		PD0
#define COL1			PD1
#define COL2			PD2
#define COL3			PD3
#define A0				PD4
#define A1				PD5
#define A2				PD6
#define ENABLE_INPUT	PD7


#define SET_LINE1_RED() 	PORTD |= _BV(A0); PORTD &= ~_BV(A1); PORTD &= ~_BV(A2)
#define SET_LINE1_GREEN() 	PORTD &= ~_BV(A0); PORTD &= ~_BV(A1); PORTD &= ~_BV(A2)
#define SET_LINE2_RED() 	PORTD &= ~_BV(A0); PORTD |= _BV(A1); PORTD &= ~_BV(A2)
#define SET_LINE2_GREEN() 	PORTD |= _BV(A0); PORTD |= _BV(A1); PORTD &= ~_BV(A2)
#define SET_LINE3_RED() 	PORTD &= ~_BV(A0); PORTD &= ~_BV(A1); PORTD |= _BV(A2)
#define SET_LINE3_GREEN() 	PORTD |= _BV(A0); PORTD &= ~_BV(A1); PORTD |= _BV(A2)

#define CLEAR_COLS()	PORTD |= _BV(COL1); PORTD |= _BV(COL2); PORTD |= _BV(COL3)
#define SET_COL1()	PORTD &= ~_BV(COL1); PORTD |= _BV(COL2); PORTD |= _BV(COL3)
#define SET_COL2()	PORTD |= _BV(COL1); PORTD &= ~_BV(COL2); PORTD |= _BV(COL3)
#define SET_COL3()	PORTD |= _BV(COL1); PORTD |= _BV(COL2); PORTD &= ~_BV(COL3)

#define EMPTY 		0
#define COMPUTER	1
#define PLAYER		2

int8_t matrix[9] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
int8_t debounced = 1, player, winner, free_leds, row, col;

void init();
void test();
void move();
void check();
void select();
void light_matrix();

void timer1_init(void)
{
    TCCR1A &= ~((1 << COM1A0) | (1 << COM1A1) |     // OC0A dezactivat
                (1 << COM1B0) | (1 << COM1B1) |     // OC0B dezactivat
                (1 << WGM10) | (1 << WGM11));       
    TCCR1B &= ~((1 << WGM12) | (1 << WGM13) | (1 << CS11) | (1 << CS12));  // mod normal
    TCCR1B |=   (1 << CS10);                        // fara prescalar
}

ISR(PCINT0_vect, ISR_BLOCK)
{
	if (player == PLAYER && debounced == 1) {
		srand((unsigned) TCNT1);
		int8_t i, j;

		// Up button
		if (bit_is_clear(PINA, SUS)) {
			if (row > 0 && matrix[(row - 1) * 3 + col] == EMPTY) {
				debounced = 0;
				matrix[row * 3 + col] = EMPTY;
				row -= 1;
				matrix[row * 3 + col] = PLAYER;
				
			} else {		
				debounced = 0;
				for (i = row - 1; i >= 0; --i)
					for (j = 0; j < 3; ++j)
						if (matrix[i * 3 + j] == EMPTY) {
							matrix[row * 3 + col] = EMPTY;
							row = i, col = j;
							matrix[row * 3 + col] = PLAYER;
						}           
			}
		    
		// Down button
		} else if (bit_is_clear(PINA, JOS)) {
        
			if (row < 2 && matrix[(row + 1) * 3 + col] == EMPTY) {
				debounced = 0;
				matrix[row * 3 + col] = EMPTY;
				row += 1;
				matrix[row * 3 + col] = PLAYER;
				
			} else {	
				debounced = 0;
				for (i = row + 1; i < 3; ++i)
					for (j = 0; j < 3; ++j)
						if (matrix[i * 3 + j] == EMPTY) {
							matrix[row * 3 + col] = EMPTY;
							row = i, col = j;
							matrix[row * 3 + col] = PLAYER;
						}
			}
        	
		// Left button
		} else if (bit_is_clear(PINA, STANGA)) {
        
			if (col > 0 && matrix[row * 3 + (col - 1)] == EMPTY) {
				debounced = 0;
				matrix[row * 3 + col] = EMPTY;
				col -= 1;
				matrix[row * 3 + col] = PLAYER;
				
			} else {
				debounced = 0;
				for (i = 0; i < 3; ++i)
					for (j = col - 1; j >= 0; --j)
						if (matrix[i * 3 + j] == EMPTY) {
							matrix[row * 3 + col] = EMPTY;
							row = i, col = j;
							matrix[row * 3 + col] = PLAYER;
						}
			}
   
		// Right button
		} else if (bit_is_clear(PINA, DREAPTA)) {
			if (col < 2 && matrix[row * 3 + (col + 1)] == EMPTY) {
				debounced = 0;
				matrix[row * 3 + col] = EMPTY;
				col += 1;
				matrix[row * 3 + col] = PLAYER;
				
			} else {	
				debounced = 0;
				for (i = 0; i < 3; ++i)
					for (j = col + 1; j < 3; ++j)
						if (matrix[i * 3 + j] == EMPTY) {
							matrix[row * 3 + col] = EMPTY;
							row = i, col = j;
							matrix[row * 3 + col] = PLAYER;
						}
			}
		
		// Select button
		} else if (bit_is_clear(PINA, SELECT)) {
			debounced = 0;
			player = COMPUTER;
			free_leds -= 1;
			check();
		}
	} else 
		debounced = 1;
}

int main(void)
{
	init();
	timer1_init();

	
	player = PLAYER;
	winner = EMPTY;
	row = 1, col = 1, free_leds = 9;

	while (1) {
		if (winner != EMPTY) {
			goto end1;
		
		} else if (free_leds <= 0) {
			goto end2;
			
		} else if (player == PLAYER) {
			matrix[row * 3 + col] = PLAYER;
		
		} else if (player == COMPUTER) {
			move();
			check();
			select();
			player = PLAYER;
		}
			
		light_matrix();
	}
	
end1:
	cli();
	if (winner == COMPUTER)
		PORTA &= ~_BV(MOD_LED_RED);
	else
		PORTA &= ~_BV(MOD_LED_GREEN);
	while (1) {
		light_matrix();
	}

end2:
	cli();
	PORTD |= _BV(USER_LED);
	while (1) {
		light_matrix();
	}

	return 0;
}

void init(void)
{
	// Port A intrare
	DDRA = 0x00;

	// Rezistentele de pull-up
	PORTA |= _BV(SUS) | _BV(JOS);
	PORTA |= _BV(STANGA) | _BV(DREAPTA);
	PORTA |= _BV(SELECT);
	PORTA |= _BV(MOD_BUTTON);
	
	// Port A LEDs
	DDRA |= _BV(USER_LED);
	DDRA |= _BV(MOD_LED_GREEN) | _BV(MOD_LED_RED);
	PORTA |= _BV(MOD_LED_GREEN) | _BV(MOD_LED_RED);
		
	// Port B iesire
	DDRD = 0xFF;
	
	// Enable input
	PORTD |= _BV(ENABLE_INPUT);
	
	// Disable column select
	CLEAR_COLS();
	
	// Activate interrupts on push_buttons
	sei();
	PCICR |= _BV(PCIE0);
	PCMSK0 |= _BV(SUS) | _BV(JOS);
	PCMSK0 |= _BV(STANGA) | _BV(DREAPTA);
	PCMSK0 |= _BV(SELECT);
}

void move(void)
{
	if (free_leds <= 0) {
		free_leds = 0;
		return;
	}

	int8_t pos = rand() % 9;
	
	while (matrix[pos] != EMPTY) {
		pos = rand() % 9;
		if (pos < 0 || pos >= 9) {
			cli();
			PORTD |= _BV(USER_LED);
		}
	}
	
	matrix[pos] = COMPUTER;
	free_leds -= 1;
}

void select(void)
{
	if (free_leds <= 0) {
		free_leds = 0;
		return;
	}
	
	int8_t i = 1;
	
	for (i = 1; i >= 0; --i)
		if (matrix[i * 3 + 1] == EMPTY) {
			row = i, col = 1;
			return;
		} else if (matrix[i * 3] == EMPTY) {
			row = i, col = 0;
			return;
		} else if (matrix[i * 3 + 2] == EMPTY) {
			row = i, col = 2;
			return;
		}
		
	i = 2;
	if (matrix[i * 3 + 1] == EMPTY) {
		row = i, col = 1;
		return;
	} else if (matrix[i * 3] == EMPTY) {
		row = i, col = 0;
		return;
	} else if (matrix[i * 3 + 2] == EMPTY) {
		row = i, col = 2;
		return;
	}
}

void zero(void)
{
	int8_t i;
	for (i = 0; i < 9; ++i)
		matrix[i] = EMPTY;
}

void check()
{
	int8_t i, j;
	
	// Check lines
	for (i = 0; i < 3; ++i)
		if (matrix[i * 3] != EMPTY && matrix[i * 3] == matrix[i * 3 + 1] && 
									  matrix[i * 3] == matrix[i * 3 + 2]) {
			winner = matrix[i * 3];
			zero();
			matrix[i * 3] = matrix[i * 3 + 1] = matrix[i * 3 + 2] = winner;
			return;
		}
		
	// Check columns
	for (j = 0; j < 3; ++j)
		if (matrix[0 * 3 + j] != EMPTY && matrix[0 * 3 + j] == matrix[1 * 3 + j] &&
										  matrix[0 * 3 + j] == matrix[2 * 3 + j]) {
			winner = matrix[0 * 3 + j];
			zero();
			matrix[0 * 3 + j] = matrix[1 * 3 + j] = matrix[2 * 3 + j] = winner;
			return;
		}
	
	// Check first diagonal
	if (matrix[0] != EMPTY && matrix[0] == matrix[4] && matrix[0] == matrix[8]) {
			winner = matrix[4];
			zero();
			matrix[4] = matrix[0] = matrix[8] = winner;
			return;
	}
	
	// Check second diagonal
	if (matrix[2] != EMPTY && matrix[2] == matrix[4]  && matrix[2] == matrix[6]) {
			winner = matrix[4];
			zero();
			matrix[4] = matrix[2] = matrix[6] = winner;
			return;
	}
}

inline void light_col(int8_t j)
{	
	if (j == 0) {
		SET_COL1();	
		
	} else if (j == 1) {
		SET_COL2();	
		
	} else {
		SET_COL3();
	}		
}

inline void light_row_player(int8_t i)
{
	if (i == 0) {
		SET_LINE1_GREEN();
		
	} else if (i == 1) {
		SET_LINE2_GREEN();
				
	} else {
		SET_LINE3_GREEN();
	}
}

inline void light_row_computer(int8_t i)
{
	if (i == 0) {
		SET_LINE1_RED();
		
	} else if (i == 1) {
		SET_LINE2_RED();
				
	} else {
		SET_LINE3_RED();
	}
}

void light_matrix(void)
{
	int8_t pos, i, j;
	for (i = 0; i < 3; ++i)
		for (j = 0; j < 3; ++j) {
			pos = i * 3 + j;
			
			if (matrix[pos] == EMPTY) {
				CLEAR_COLS();
				continue;
				
			} else if (matrix[pos] == PLAYER) {
				light_row_player(i);
				
			} else if (matrix[pos] == COMPUTER) {
				light_row_computer(i);
			}
			light_col(j);
			
			_delay_us(100);
		}
}

void test(void)
{
	while (1) {
		// LINE 1
		SET_LINE1_RED();
		SET_COL1();
		_delay_ms(500);
		
		SET_COL2();
		_delay_ms(500);
		
		SET_COL3();
		_delay_ms(500);
		
		// LINE 2
		SET_LINE2_RED();
		SET_COL1();
		_delay_ms(500);
		
		SET_COL2();
		_delay_ms(500);
		
		SET_COL3();
		_delay_ms(500);
		
		// LINE 3
		SET_LINE3_RED();
		SET_COL1();
		_delay_ms(500);
		
		SET_COL2();
		_delay_ms(500);
		
		SET_COL3();
		_delay_ms(500);
		
		
		// LINE 1
		SET_LINE1_GREEN();
		SET_COL1();
		_delay_ms(500);
		
		SET_COL2();
		_delay_ms(500);
		
		SET_COL3();
		_delay_ms(500);
		
		// LINE 2
		SET_LINE2_GREEN();
		SET_COL1();
		_delay_ms(500);
		
		SET_COL2();
		_delay_ms(500);
		
		SET_COL3();
		_delay_ms(500);
		
		// LINE 3
		SET_LINE3_GREEN();
		SET_COL1();
		_delay_ms(500);
		
		SET_COL2();
		_delay_ms(500);
		
		SET_COL3();
		_delay_ms(500);
	}
}
